//---------------------------------------------------------------------------
//
//  Module:   testapp.c
//
//  Purpose:
//     Exercises the VDISKIOD virtual device driver.
//
//
//  Description of functions:
//     See below.
//
//
//  Development Team:
//     Bryan A. Woodruff
//
//  History:   Date       Author      Comment
//             10/ 5/92   BryanW      Added this comment block.
//
//---------------------------------------------------------------------------
//
//  Written by Microsoft Product Support Service, Windows Developer Support.
//  Copyright (c) 1992 Microsoft Corporation.  All Rights Reserved.
//
//---------------------------------------------------------------------------

#define GLOBALDEFS  //  Global definitions, this module only

#include "testapp.h"

//---------------------------------------------------------------------------
//  int PASCAL WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance,
//                      LPSTR lpCmdLine, int nCmdShow )
//
//  Description:
//     This is the main window entry point.
//
//  Parameters:
//     HINSTANCE hInstance
//        handle to this instance
//
//     HINSTANCE hPrevInstance
//        handle to previous instance
//
//     LPSTR lpCmdLine
//        pointer to command line
//
//     int nCmdShow
//        defines the mode to show the window
//
//  History:   Date       Author      Comment
//             10/ 5/92   BryanW      Added this comment block.
//
//---------------------------------------------------------------------------

int PASCAL WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance,
                    LPSTR lpCmdLine, int nCmdShow )
{
   MSG  msg ;

   if (!hPrevInstance)
      if (!InitApplication( hInstance ))
         return ( FALSE ) ;

   if (!InitInstance( hInstance, nCmdShow ))
      return ( FALSE ) ;

   while (GetMessage( &msg, NULL, NULL, NULL ))
   {
      TranslateMessage( &msg ) ;
      DispatchMessage( &msg ) ;
   }
   return ( msg.wParam ) ;

} // end of WinMain()

//---------------------------------------------------------------------------
//  BOOL InitApplication( HINSTANCE hInstance )
//
//  Description:
//     Application specific initialization.  Performed only once
//     when first instance of application is started.
//
//  Parameters:
//     HINSTANCE hInstance
//        handle to this instance
//
//  History:   Date       Author      Comment
//             10/ 5/92   BryanW      Added this comment block
//
//---------------------------------------------------------------------------

BOOL InitApplication( HINSTANCE hInstance )
{
   WNDCLASS  wndclass ;

   wndclass.style =         NULL ;
   wndclass.lpfnWndProc =   MainWndProc ;
   wndclass.cbClsExtra =    0 ;
   wndclass.cbWndExtra =    0 ;
   wndclass.hInstance =     hInstance ;
   wndclass.hIcon =         LoadIcon( NULL, IDI_APPLICATION ) ;
   wndclass.hCursor =       LoadCursor( NULL, IDC_ARROW ) ;
   wndclass.hbrBackground = GetStockObject( WHITE_BRUSH ) ;
   wndclass.lpszMenuName =  "GenericMenu" ;
   wndclass.lpszClassName = "GenericWClass" ;

   return ( RegisterClass( &wndclass ) ) ;

} // end of InitApplication()

//---------------------------------------------------------------------------
//  BOOL InitInstance( HINSTANCE hInstance, int nCmdShow )
//
//  Description:
//     Handles the initialization for each instance of
//     this applicaton.
//
//  Parameters:
//     HINSTANCE hInstance
//        handle to this instance
//
//     int nCmdShow
//        command to determine how to show this window
//
//  History:   Date       Author      Comment
//             10/ 5/92   BryanW      Added this comment block.
//
//---------------------------------------------------------------------------

BOOL InitInstance( HINSTANCE hInstance, int nCmdShow )
{
   HWND  hWnd ;

   hInst = hInstance ;

   hWnd = CreateWindow( "GenericWClass", "Generic Sample Application",
                        WS_OVERLAPPEDWINDOW, CW_USEDEFAULT, CW_USEDEFAULT,
                        CW_USEDEFAULT, CW_USEDEFAULT, NULL, NULL,
                        hInstance, NULL ) ;

   if (!hWnd)
      return ( FALSE ) ;

   ShowWindow( hWnd, nCmdShow ) ;
   UpdateWindow( hWnd ) ;
   return ( TRUE ) ;

} // end of InitInstance()

//---------------------------------------------------------------------------
//  LRESULT FAR PASCAL MainWndProc( HWND hWnd, UINT uMsg,
//                                  WPARAM wParam, LPARAM lParam )
//
//  Description:
//
//
//  Parameters:
//     HWND hWnd
//        handle of this window
//
//     UINT uMsg
//        message for processing
//
//     WPARAM wParam
//        word parameter for message
//
//     LPARAM lParam
//        dword parameter for message
//
//  History:   Date       Author      Comment
//             10/ 5/92   BryanW      Added this comment block.
//
//---------------------------------------------------------------------------

LRESULT FAR PASCAL MainWndProc( HWND hWnd, UINT uMsg,
                                WPARAM wParam, LPARAM lParam )
{
   switch (uMsg)
   {
      case WM_NOTIFICATION:
         switch (LOWORD( lParam ))
         {
            case NOTIFY_ReadBufferFull:
            {
               char  szData[ 1024 ] ;
               HDC   hDC ;
               RECT  rcText ;
               WORD  wSize ;

               GetClientRect( hWnd, &rcText ) ;
               hDC = GetDC( hWnd ) ;
               while (wSize = ReadData( szData ))
               {
                  FillRect( hDC, &rcText, GetStockObject( WHITE_BRUSH ) ) ;
                  DrawText( hDC, szData, wSize, &rcText,
                            DT_WORDBREAK | DT_NOPREFIX ) ;
               }
               ReleaseDC( hWnd, hDC ) ;
            }
            break ;

            case NOTIFY_ErrorStatus:
            {
               char  szBuf[ 80 ] ;

               wsprintf( szBuf, "DOS Error #%d", HIWORD( lParam ) ) ;
               MessageBox( hWnd, szBuf, "Read Error!",
                           MB_OK | MB_ICONEXCLAMATION ) ;
            }
            break ;

            case NOTIFY_ReadComplete:
               MessageBox( hWnd, "Read Completed!", "Important!", MB_OK ) ;
         }
         break ;

      case WM_COMMAND:
         switch (wParam)
         {
            case IDM_STARTIO:
               StartIO( hWnd, "test.txt" ) ;
               break ;

            case IDM_ABOUT:
            {
               FARPROC  lpProcAbout ;

               lpProcAbout = MakeProcInstance( AboutDlgProc, hInst ) ;
               DialogBox( hInst, "AboutBox", hWnd, lpProcAbout ) ;
               FreeProcInstance( lpProcAbout ) ;
            }
            break ;
         }
         break ;

      case WM_DESTROY:
         PostQuitMessage( 0 ) ;
         break ;

      default:
         return ( DefWindowProc( hWnd, uMsg, wParam, lParam ) ) ;
   }
   return ( NULL ) ;

} // end of MainWndProc()

/************************************************************************
 *  BOOL FAR PASCAL AboutDlgProc( HWND hDlg, UINT uMsg,
 *                                WPARAM wParam, LPARAM lParam )
 *
 *  Description:
 *     About dialog box procedure.  This procedure handles all messages
 *     for the about dialog box.
 *
 *  Comments:
 *      5/28/91  baw  Cleaned up the code.
 *
 ************************************************************************/

//---------------------------------------------------------------------------
//  BOOL FAR PASCAL AboutDlgProc( HWND hDlg, UINT uMsg,
//                                WPARAM wParam, LPARAM lParam )
//
//  Description:
//     Handles the message processing for the dialog box.
//
//  Parameters:
//     HWND hDlg
//        handle to dialog box
//
//     UINT uMsg
//        message to process
//
//     WPARAM wParam
//        word parameter for message
//
//     LPARAM lParam
//        dword parameter for message
//
//  History:   Date       Author      Comment
//             10/ 5/92   BryanW      Added this comment block.
//
//---------------------------------------------------------------------------

BOOL FAR PASCAL AboutDlgProc( HWND hDlg, UINT uMsg,
                              WPARAM wParam, LPARAM lParam )
{
   switch (uMsg)
   {
      case WM_INITDIALOG:
         return ( TRUE ) ;

      case WM_COMMAND:
         switch (wParam)
         {
            case IDOK:
            case IDCANCEL:
               EndDialog( hDlg, TRUE ) ;
               return ( TRUE ) ;
         }
         break ;
   }
   return ( FALSE ) ;

} // end of AboutDlgProc()

//---------------------------------------------------------------------------
//  BOOL NEAR StartIO( HWND hWnd, LPSTR lpFileName )
//
//  Description:
//     Initializes VDISKIOD and starts the disk I/O of the
//     given file.
//     
//  Parameters:
//     HWND hWnd
//        handle to window to receive notifications
//
//     LPSTR lpFileName
//        pointer to filename
//
//  History:   Date       Author      Comment
//             10/ 5/92   BryanW      Wrote it.
//
//---------------------------------------------------------------------------

BOOL NEAR StartIO( HWND hWnd, LPSTR lpFileName )
{
   BOOL     fRetVal = FALSE ;
   DWORD    dwVxDEntry ;
   FARPROC  lpPostMessage ;

   lpPostMessage = GetProcAddress( GetModuleHandle( "USER" ),
                                   MAKEINTRESOURCE( 110 ) ) ;

   _asm
   {
      push  es
      push  si
      push  di

      mov   ax, 0x1684
      mov   bx, VDISKIOD_Device_ID
      int   0x2f
      mov   word ptr dwVxDEntry, di
      mov   word ptr dwVxDEntry + 2, es
      mov   ax, es
      or    ax, di
      jz    SIO_Out

      mov   dx, 1                         ; VDISKIOD_Init
      les   bx, lpPostMessage             ; ES:BX <- ptr of PostMessage
      call  dword ptr dwVxDEntry          ; call VDISKIOD

      mov   dx, 2                         ; VDISKIOD_Start_Disk_IO
      mov   ax, WM_NOTIFICATION           ; message for callback
      mov   bx, hWnd                      ; handle to callback window
      les   si, lpFileName                ; ES:SI <- pointer to filename
      call  dword ptr dwVxDEntry          ; call VDISKIOD

      mov   fRetVal, TRUE

   SIO_Out:
      pop   di
      pop   si
      pop   es

   }

   return ( fRetVal ) ;

} // end of StartIO()

//------------------------------------------------------------------------
//  WORD NEAR ReadData( LPSTR lpBuf )
//
//  Description:
//     Calls the VxD to copy a read buffer to the given
//     local buffer.
//
//  Parameters:
//     LPSTR lpBuf
//        pointer to local buffer
//
//  History:   Date       Author      Comment
//             10/ 5/92   BryanW      Wrote it.
//
//------------------------------------------------------------------------

WORD NEAR ReadData( LPSTR lpBuf )
{

   DWORD  dwVxDEntry ;
   WORD   wSize = 0 ;

   _asm
   {
      push  es
      push  si
      push  di

      mov   ax, 0x1684
      mov   bx, VDISKIOD_Device_ID
      int   0x2f
      mov   word ptr dwVxDEntry, di
      mov   word ptr dwVxDEntry + 2, es
      mov   ax, es
      or    ax, di
      jz    RD_Out

      push  ds
      mov   dx, 3                         ; VDISKIOD_Start_Disk_IO
      lds   di, lpBuf                     ; DS:DI <- pointer to buffer
      call  dword ptr dwVxDEntry          ; call VDISKIOD
      pop   ds

      mov   wSize, ax

   RD_Out:
      pop   di
      pop   si
      pop   es

   }

   return ( wSize ) ;

} // end of ReadData()

//---------------------------------------------------------------------------
//  End of File: testapp.c
//---------------------------------------------------------------------------

