;---------------------------------------------------------------------------
;
;  Module:   vdiskiod.asm
;
;  Purpose:
;     A sample VxD to demonstrate background disk I/O initiated
;     through an API call from a Windows application.
;
;  Development Team:
;     Bryan A. Woodruff
;     Brian Lieuellen
;
;  History:   Date       Author      Comment
;             10/16/91   BrianLie    Original sources for disk I/O.
;             10/ 4/92   BryanW      Created the sample.  Added features
;                                    and notifications to WinApp.
;
;---------------------------------------------------------------------------
;
;  Written by Windows Premier Developer Support.
;  Copyright (c) 1992 Microsoft Corporation.  All Rights Reserved.
;
;---------------------------------------------------------------------------

        .386p

;***************************************************************************
;                             I N C L U D E S
;***************************************************************************

        .XLIST
        INCLUDE VMM.Inc
        INCLUDE DOSMgr.Inc
        INCLUDE Debug.Inc
        .LIST

        Create_VDISKIOD_Service_Table EQU 1       
        INCLUDE VDISKIOD.Inc

;***************************************************************************
;                            E Q U A T E S
;***************************************************************************

DOS_OpenFile                    equ     3d00h
DOS_CloseFile                   equ     3e00h
DOS_ReadFile                    equ     3f00h
DOS_WriteFile                   equ     4000h

DOS_GetExtendedError            equ     5900h

OPEN_Access_ReadOnly            equ     0000h
OPEN_Access_WriteOnly           equ     0001h
OPEN_Access_ReadWrite           equ     0002h
OPEN_Share_DenyReadWrite        equ     0010h
OPEN_Share_DenyWrite            equ     0020h
OPEN_Share_DenyRead             equ     0030h
OPEN_Share_DenyNone             equ     0040h

STATE_Idle                      equ     0ffffh
STATE_OpenFile                  equ     0
STATE_ReadFile                  equ     1
STATE_WaitForRoom               equ     2
STATE_CloseFile                 equ     3
STATE_ReportError               equ     4

NOTIFY_ReadBufferFull           equ     0
NOTIFY_ReadComplete             equ     1
NOTIFY_ErrorStatus              equ     2

TIMEOUT_IO_Slice                equ     100             ; 100 ms timeslice
MAX_PathLen                     equ     164
MAX_BlockSize                   equ     1024
MAX_ReadBuffers                 equ     6

;***************************************************************************
;              V I R T U A L  D E V I C E  D E C L A R A T I O N
;***************************************************************************

Declare_Virtual_Device VDISKIOD, 3, 0ah, VDISKIOD_Control,\
                       VDISKIOD_Device_ID, <Int13_Init_Order + 1>,,\
                       VDISKIOD_PM_API_Proc

;***************************************************************************
;		   I N I T I A L I Z A T I O N	 D A T A
;***************************************************************************

VxD_IDATA_SEG
VxD_IDATA_ENDS

;***************************************************************************
;                    D E V I C E  F I X E D  D A T A
;***************************************************************************

VxD_LOCKED_DATA_SEG

NotifySent              dw      0

Current_PSP             dw      0
Win386_PSP              dw      0
TimeoutHandle           dd      0
FileHandle              dw      0
ReadPos_VxD             dw      0
ReadPos_App             dw      0

LastError               dw      0
NotifyWnd               dw      0
NotifyMsg               dw      0
lpPostMessage           dd      0

szFileName              db      MAX_PathLen dup (?)
IO_Buffer               db      MAX_ReadBuffers * (MAX_BlockSize + 2) dup (?)

CurrentState            dd      0                       ; for FSM states
ErrorState              dd      0

OldInt13Proc            dd      0
Inside_Int13            dd      0

VxD_LOCKED_DATA_ENDS

;***************************************************************************
;                          D E V I C E  D A T A
;***************************************************************************

VxD_DATA_SEG

VDISKIOD_Call_Table LABEL DWORD
        dd      OFFSET32        VDISKIOD_PM_API_Get_Version
        dd      OFFSET32        VDISKIOD_PM_API_Init
        dd      OFFSET32        VDISKIOD_PM_API_Start_Disk_IO
        dd      OFFSET32        VDISKIOD_PM_API_Get_Block

Max_VDISKIOD_Service    equ     ($ - VDISKIOD_Call_Table) / 4

VDISKIOD_State_Handlers LABEL DWORD
        dd      OFFSET32        VDISKIOD_STATE_OpenFile
        dd      OFFSET32        VDISKIOD_STATE_ReadFile
        dd      OFFSET32        VDISKIOD_STATE_WaitForRoom
        dd      OFFSET32        VDISKIOD_STATE_CloseFile
        dd      OFFSET32        VDISKIOD_STATE_ReportError

Max_STATE_HANDLERS      equ     5


VxD_DATA_ENDS

;***************************************************************************
;                 R E A L   M O D E   I N I T   C O D E
;***************************************************************************

;===========================================================================
;
;   VDISKIOD_Real_Mode_Init:
;
;      Checks for duplicate VxD ID and fails to load if so...
;
;===========================================================================

VxD_REAL_INIT_SEG

BeginProc VDISKIOD_Real_Mode_Init

        cmp     ax, 30ah                                ; only load under
        jne     short MR_FailLoad                       ; Windows 3.10

        test    bx, Duplicate_Device_ID
        jnz     short MR_FailLoad

        mov     ax, Device_Load_Ok
        jmp     short MR_Exit

MR_FailLoad:
        mov     ax, Abort_Device_Load + No_Fail_Message

MR_Exit:
        xor     bx, bx
        xor     si, si
        xor     edx, edx

        ret

EndProc VDISKIOD_Real_Mode_Init
   
VxD_REAL_INIT_ENDS

;***************************************************************************
;                 I N I T I A L I Z A T I O N  C O D E
;***************************************************************************

VxD_ICODE_SEG

;===========================================================================
;
;   VDISKIOD_Sys_Critical_Init:
;      Hooks the V86 INT 13h handler to watch when INT 13h is
;      entered.
;
;===========================================================================

BeginProc VDISKIOD_Sys_Critical_Init

        push    eax
        push    esi

        mov     eax, 13h
        mov     esi, OFFSET32 VDISKIOD_Int13_Watch
        VMMCall Hook_V86_Int_Chain

        pop     esi
        pop     eax

        clc
        ret

EndProc VDISKIOD_Sys_Critical_Init

;===========================================================================
;
;   VDISKIOD_Device_Init:
;      Grabs Win386's PSP for use with file operations.
;
;===========================================================================

BeginProc VDISKIOD_Device_Init

        VMMCall Get_PSP_Segment                 ; Save Win386's PSP for
        mov     Win386_PSP, ax                  ; our own use.

        clc
        ret

EndProc VDISKIOD_Device_Init


VxD_ICODE_ENDS


VxD_CODE_SEG

;***************************************************************************
;                         E X P O R T E D  A P I 
;***************************************************************************

;===========================================================================
;
;   VDISKIOD_PM_API_Proc
;
;      This is the exported API procedure callable from the System VM.
;      An application used INT 2Fh, AX=1684h, BX=Device_ID to
;      get a callback address.  When the callback address is
;      called, we eventually end up here.
;
;===========================================================================

BeginProc VDISKIOD_PM_API_Proc

        Trace_Out "In VDISKIOD_PM_API_Proc"

        movzx   eax, [ebp.Client_DX]                    ; function in DX
        cmp     eax, Max_VDISKIOD_Service
        jae     SHORT VDISKIOD_Call_Bad
        and     [ebp.Client_EFLAGS], NOT CF_Mask        ; clear carry
        call    VDISKIOD_Call_Table[ eax * 4 ]          ; call service
        ret

VDISKIOD_Call_Bad:
IFDEF DEBUG
        Debug_Out "VDISKIOD ERROR: Invalid function #EAX on VDISKIOD_PM_API_Proc"
ENDIF
        mov     [ebp.Client_EAX], 0FFFFFFFFh

VDISKIOD_PM_API_Failed:
        or      [ebp.Client_EFLAGS], CF_Mask            ; set carry
        ret

EndProc VDISKIOD_PM_API_Proc

;===========================================================================
;
;   VDISKIOD_PM_API_Get_Version
;
;      Version call for PM API entry point
;
;   ENTRY:
;      Client_DX = 0
;
;   EXIT:
;      if Client_EFLAGS.CF_Mask clear
;         Client_AX = Version
;      else
;         VDISKIOD device is not installed
;
;   USES:       FLAGS, EAX
;
;===========================================================================

BeginProc VDISKIOD_PM_API_Get_Version

        Trace_Out "In VDISKIOD_PM_API_Get_Version"

        VxDCall VDISKIOD_Get_Version
        jc      VDISKIOD_PM_API_Failed
        mov     [ebp.Client_AX], ax

        ret

EndProc VDISKIOD_PM_API_Get_Version

;===========================================================================
;
;   VDISKIOD_PM_API_Init
;
;      Initializes the lpPostMessage information.
;
;   ENTRY:
;      Client_DX = 1
;      Client_ES:Client_BX = Pointer to PostMessage() function
;
;   EXIT:
;      if Client_EFLAGS.CF_Mask clear
;         initialization complete
;         Client_AX contains block size
;      else
;         something unexpected happened
;
;   USES:       FLAGS, EAX
;
;===========================================================================

BeginProc VDISKIOD_PM_API_Init

        Trace_Out "In VDISKIOD_PM_API_Init"

        movzx   eax, [ebp.Client_BX]
        mov     WORD PTR lpPostMessage, ax
        movzx   eax, [ebp.Client_ES]
        mov     WORD PTR lpPostMessage + 2, ax
        mov     [ebp.Client_AX], MAX_BlockSize + 2

        ret

EndProc VDISKIOD_PM_API_Init

;===========================================================================
;
;   VDISKIOD_PM_API_Start_Disk_IO
;
;      Starts the background disk I/O by calling the appropriate
;      VDISKIOD service.
;
;   ENTRY:
;      Client_DX = 2
;      Client_AX = Message ID to use with notifications
;      Client_BX = handle to notification window
;      Client_ES:Client_SI = Pointer to null terminated filename
;
;   EXIT:
;      if Client_EFLAGS.CF_Mask clear
;         IO was started successfully
;      else
;         something failed
;
;   USES: FLAGS, EAX
;
;===========================================================================

BeginProc VDISKIOD_PM_API_Start_Disk_IO

        Trace_Out "In VDISKIOD_PM_API_Start_Disk_IO"

        push    eax
        push    edx
        push    esi

        movzx   eax, [ebp.Client_AX]
        movzx   edx, [ebp.Client_BX]
        Client_Ptr_Flat esi, ES, SI
        VxDCall VDISKIOD_Start_Disk_IO

        pop     esi
        pop     edx
        pop     eax

        jc      VDISKIOD_PM_API_Failed
        ret

EndProc VDISKIOD_PM_API_Start_Disk_IO

;===========================================================================
;
;   VDISKIOD_PM_API_Get_Block
;
;      Gets a block from the read buffer or returns with error code.
;
;   ENTRY:
;      Client_DX = 3
;      Client_DS:Client_DI = Pointer to buffer to fill
;
;   EXIT:
;      if Client_EFLAGS.CF_Mask clear
;         transfer completed and
;         Client_AX contains the size of the transfer
;      else Client_EFLAGS.CF_Mask set
;         Client_AX contains error code
;
;   USES: FLAGS, EAX
;
;===========================================================================

BeginProc VDISKIOD_PM_API_Get_Block

        Trace_Out "In VDISKIOD_PM_Get_Block"

        push    eax
        push    edi

        Client_Ptr_Flat edi, DS, DI
        VxDCall VDISKIOD_Get_Block
        mov     [ebp.Client_AX], ax

        pop     edi
        pop     eax

        jc      VDISKIOD_PM_API_Failed
        ret

EndProc VDISKIOD_PM_API_Get_Block

;===========================================================================
;
;   VDISKIOD_Get_Version
;
;      Get VDISKIOD device version
;
;   EXIT:
;      if carry clear
;         EAX = Version
;      else
;         VDISKIOD device is not installed
;
;   USES:       FLAGS, EAX
;
;===========================================================================

BeginProc VDISKIOD_Get_Version, Service

        mov     eax, 30Ah
        clc
        ret

EndProc VDISKIOD_Get_Version

;===========================================================================
;
;   VDISKIOD_Start_Disk_IO
;
;      Starts the background disk I/O using the given parameters...
;      see below for a description of register usage.
;
;   ENTRY:
;      EAX = message to use for notifications to Windows app
;      EDX = handle to application's window
;      ESI = pointer to szFileName
;
;   EXIT:
;      if carry clear
;         IO started successfully
;      else
;         something failed
;
;   USES:
;      FLAGS
;
;===========================================================================

BeginProc VDISKIOD_Start_Disk_IO, Service

        push    edi
        push    esi
        push    edx
        push    eax

        mov     NotifyWnd, dx
        mov     NotifyMsg, ax

        ; copy string from VM's area to our own

        mov     edi, OFFSET32 szFileName
VSDI_Copy:
        lodsb
        stosb
        cmp     al, 0
        jne     SHORT VSDI_Copy

        mov     CurrentState, STATE_OpenFile
        mov     eax, TIMEOUT_IO_Slice
        xor     edx, edx
        mov     esi, OFFSET32 VDISKIOD_TimeOut_Callback
        VMMCall Set_Global_Time_Out

        pop     eax
        pop     edx
        pop     esi
        pop     edi

        clc
        ret

EndProc VDISKIOD_Start_Disk_IO

;===========================================================================
;
;   VDISKIOD_TimeOut_Callback
;
;      This is the finite state machine entry point that will
;      dispatch to the state handler for the current state.
;
;   ENTRY:
;      EBX = current VM
;      EBP = client register structure
;
;   EXIT:
;      none
;
;   USES:
;      FLAGS
;
;===========================================================================

BeginProc VDISKIOD_TimeOut_Callback, High_Freq

        push    eax
        push    edx
        push    esi

        mov     eax, CurrentState                       ; get current state
        cmp     eax, STATE_Idle                         ; Q: idle?
        jae     SHORT VDISKIOD_IC_Exit                  ;    Y: no work to do
        cmp     eax, Max_STATE_HANDLERS                 ; Q: invalid?
        jae     SHORT VDISKIOD_IC_Bad_State             ;    Y: just go idle

        call    VDISKIOD_State_Handlers[ eax * 4 ]      ; call state handler

        mov     eax, TIMEOUT_IO_Slice                   ; reschedule timeout
        xor     edx, edx
        mov     esi, OFFSET32 VDISKIOD_Timeout_Callback
        VMMCall Set_Global_Time_Out
        jmp     SHORT VDISKIOD_IC_Exit

VDISKIOD_IC_Bad_State:
        mov     CurrentState, STATE_Idle                ; just force to idle

VDISKIOD_IC_Exit:
        pop     esi
        pop     edx
        pop     eax

        ret

EndProc VDISKIOD_TimeOut_Callback

;===========================================================================
;
;   VDISKIOD_STATE_OpenFile
;
;      Opens the file is I/O is valid.
;
;   USES:
;      EAX, ECX, EDX, FLAGS
;
;===========================================================================

BeginProc VDISKIOD_STATE_OpenFile, High_Freq

        Trace_Out "STATE: OpenFile"

        call    VDISKIOD_Test_IO_Valid                  ; Q: can we I/O?
        jc      VSOF_JustRet                            ;    N: just wait

        mov     ecx, (Block_Svc_Ints OR Block_Enable_Ints)
        VMMCall Begin_Critical_Section

        call    VDISKIOD_Switch_PSP_To_Win386

        mov     edx, OFFSET32 szFileName
        mov     eax, DOS_OpenFile OR OPEN_Access_ReadWrite
        VxDInt  21h
        jc      SHORT DOS_Error
        mov     FileHandle, ax
        mov     CurrentState, STATE_ReadFile
        jmp     SHORT VSOF_Exit

        ; Second entry point !!!

DOS_Error:
        mov     eax, DOS_GetExtendedError
        VxDInt  21h
        mov     LastError, ax

        Debug_Out "DOS call failed! DOS extended error = #AX"

        mov     eax, CurrentState
        mov     ErrorState, eax
        mov     CurrentState, STATE_ReportError

VSOF_Exit:
        call    VDISKIOD_Switch_PSP_To_Current
        VMMCall End_Critical_Section

VSOF_JustRet:
        ret

EndProc VDISKIOD_STATE_OpenFile

;===========================================================================
;
;   VDISKIOD_STATE_ReadFile
;
;      Checks the I/O state.  If I/O is valid, checks for room
;      in the read buffer and performs I/O using DOS.
;
;   USES:
;      EAX, ECX, EDX, FLAGS
;
;===========================================================================

BeginProc VDISKIOD_STATE_ReadFile, High_Freq

        Trace_Out "STATE: ReadFile"

        call    VDISKIOD_Room_In_ReadBuffer             ; Q: Is there room?
        jnc     SHORT VSRF_RoomForReading               ;    Y: go ahead
        mov     CurrentState, STATE_WaitForRoom         ;    N: wait for room
        ret

VSRF_RoomForReading:
        call    VDISKIOD_Test_IO_Valid                  ; Q: can we I/O?
        jc      SHORT VSRF_JustRet                      ;    N: just wait

        mov     ecx, (Block_Svc_Ints OR Block_Enable_Ints)
        VMMCall Begin_Critical_Section

        call    VDISKIOD_Switch_PSP_To_Win386

        push    ebx
        movzx   ebx, FileHandle
        mov     ecx, MAX_BlockSize 
        movzx   eax, ReadPos_VxD                        ; get current position
        mov     edx, MAX_BlockSize + 2                  ; and compute offset
        mul     edx                                     ; into the IO buffer
        add     eax, OFFSET32 IO_Buffer + 2              
        mov     edx, eax                                
        mov     eax, DOS_ReadFile
        VxDInt  21h
        pop     ebx
        jc      DOS_Error
        mov     word ptr [edx - 2], ax                  ; save size

        movzx   eax, ReadPos_VxD                        ; increment buffer
        call    VDISKIOD_Increment_ReadPos              ; read position 
        mov     ReadPos_VxD, ax

        cmp     word ptr [edx - 2], MAX_BlockSize           ; check for EOF
        je      SHORT VSRF_Exit

        mov     CurrentState, STATE_CloseFile

VSRF_Exit:
        call    VDISKIOD_Switch_PSP_To_Current
        VMMCall End_Critical_Section

VSRF_JustRet:
        ret

EndProc VDISKIOD_STATE_ReadFile

;===========================================================================
;
;   VDISKIOD_STATE_Room_In_ReadBuffer
;
;      Checks for room in the read IO buffer.  If no room, returns
;      with carry set, otherwise carry is clear.
;
;   USES:
;      EAX, FLAGS
;
;===========================================================================

BeginProc VDISKIOD_Room_In_ReadBuffer, High_Freq

        movzx   eax, ReadPos_VxD
        call    VDISKIOD_Increment_ReadPos
        cmp     ax, ReadPos_App
        je      SHORT VRIRB_NoRoom
        clc
        ret

VRIRB_NoRoom:
        stc
        ret

EndProc VDISKIOD_Room_In_ReadBuffer

;===========================================================================
;
;   VDISKIOD_Increment_ReadPos
;
;      Increments EAX to the next read position in the buffer.
;
;   ENTRY:
;      EAX: contains the the current read position
;
;   USES:
;      EAX, FLAGS
;
;===========================================================================

BeginProc VDISKIOD_Increment_ReadPos, High_Freq

        inc     eax
        cmp     eax, MAX_ReadBuffers
        jb      SHORT VIRO_Exit
        xor     eax, eax

VIRO_Exit:
        ret

EndProc VDISKIOD_Increment_ReadPos

;===========================================================================
;
;   VDISKIOD_NotifyApp
;
;      This routine notifies the Windows application through a
;      call to the PostMessage() API.
;
;   ENTRY:
;      EDX: contains the lParam of the message
;
;   USES:
;      FLAGS
;
;===========================================================================

BeginProc VDISKIOD_NotifyApp, High_Freq

        push    ebx
        mov     eax, High_Pri_Device_Boost
        VMMCall Get_Sys_VM_Handle
        mov     ecx, PEF_Wait_For_STI OR PEF_Wait_Not_Crit or PEF_Always_Sched
        mov     esi, OFFSET32 VDISKIOD_PostEvent
        xor     edi, edi
        VMMCall Call_Priority_VM_Event
        pop     ebx

VSNA_Exit:
        ret

EndProc VDISKIOD_NotifyApp

;===========================================================================
;
;   VDISKIOD_STATE_WaitForRoom
;
;      Checks for room in the read IO buffer.  If no room, will notify
;      application of buffer full if not already notified.
;
;   USES:
;      ECX, EDX, FLAGS
;
;===========================================================================

BeginProc VDISKIOD_STATE_WaitForRoom, High_Freq

        Trace_Out "STATE: WaitForRoom"

        call    VDISKIOD_Room_In_ReadBuffer             ; Q: Is there room?
        jc      SHORT VSWFR_Notify                      ;    N: notify app
        mov     CurrentState, STATE_ReadFile            ; continue reading
        ret

VSWFR_Notify:
        mov     ecx, NOTIFY_ReadBufferFull
        bts     NotifySent, cx                          ; Q: Already sent?
        jc      SHORT VSWFR_Exit                        ;    Y: just exit
        mov     edx, 1
        shl     edx, cl
        call    VDISKIOD_NotifyApp

VSWFR_Exit:
        ret

EndProc VDISKIOD_STATE_WaitForRoom

;===========================================================================
;
;   VDISKIOD_STATE_CloseFile
;
;      Closes the file handle when I/O is valid.
;
;   USES:
;      EAX, ECX, FLAGS
;
;===========================================================================

BeginProc VDISKIOD_STATE_CloseFile, High_Freq

        Trace_Out "STATE: CloseFile"

        call    VDISKIOD_Test_IO_Valid                  ; test if we can I/O
        jc      VSCF_Exit                               ; if not just wait

        mov     ecx, (Block_Svc_Ints OR Block_Enable_Ints)
        VMMCall Begin_Critical_Section

        call    VDISKIOD_Switch_PSP_To_Win386

        push    ebx
        movzx   ebx, FileHandle
        mov     eax, DOS_CloseFile
        VxDInt  21h
        pop     ebx
        jc      DOS_Error

        mov     CurrentState, STATE_Idle

        call    VDISKIOD_Switch_PSP_To_Current
        VMMCall End_Critical_Section

VSCF_NotifyComplete:
        movzx   eax, ReadPos_App
        cmp     ax, ReadPos_VxD
        je      SHORT VSCF_NoData

        ; ensure that tha app gets notification for
        ; any of the remaining data buffers

        mov     ecx, NOTIFY_ReadBufferFull
        bts     NotifySent, cx                          ; Q: Already sent?
        jc      SHORT VSCF_NoData                       ;    Y: just exit
        mov     edx, 1
        shl     edx, cl
        call    VDISKIOD_NotifyApp


VSCF_NoData:
        mov     ecx, NOTIFY_ReadComplete
        mov     edx, 1
        shl     edx, cl
        call    VDISKIOD_NotifyApp

VSCF_Exit:
        ret

EndProc VDISKIOD_STATE_CloseFile

;===========================================================================
;
;   VDISKIOD_STATE_ReportError
;
;      Reports an error to the application, if not already notified.
;      Resumes with the next logical state in the FSM.
;
;   USES:
;      EAX, ECX, EDX, FLAGS
;
;===========================================================================

BeginProc VDISKIOD_STATE_ReportError, High_Freq

        Trace_Out "STATE: ReportError"

        mov     ecx, NOTIFY_ErrorStatus
        mov     edx, 1
        shl     edx, cl
        movzx   eax, LastError
        shl     eax, 16
        or      edx, eax
        call    VDISKIOD_NotifyApp

VSRE_SetState:
        mov     eax, ErrorState

        cmp     eax, STATE_ReadFile
        jne     SHORT VSRE_NotRead
        mov     eax, STATE_CloseFile
        jmp     SHORT VSRE_Exit

VSRE_NotRead:     
        cmp     eax, STATE_CloseFile                    ; Q: Closing file?
        jne     SHORT VSRE_NotClose                     ;   N: just go idle
        jmp     VSCF_NotifyComplete                     ;   Y: complete

VSRE_NotClose:
        mov     eax, STATE_Idle

VSRE_Exit:
        mov     CurrentState, eax
        ret

EndProc VDISKIOD_STATE_ReportError

;===========================================================================
;
;   VDISKIOD_Test_IO_Valid
;
;      This routine will determine if I/O with DOS is valid.
;
;      Borrowed from PAGESWAP.ASM.
;
;   EXIT:
;      if carry set
;         I/O is not valid
;      else
;         carry is clear and I/O can be performed
;
;   USES:
;      FLAGS
;
;===========================================================================

BeginProc VDISKIOD_Test_IO_Valid, High_Freq

        VxDCall DOSMGR_Get_DOS_Crit_Status
        jne     SHORT VTIV_Not_Valid

        push    eax
        VxDCall DOSMGR_Get_IndosPtr
        cmp     WORD PTR ds:[eax], 0
        pop     eax
        jne     SHORT VTIV_Not_Valid
        cmp     Inside_Int13, 0
        jne     SHORT VTIV_Not_Valid

VTIV_Valid:
        clc
        ret

VTIV_Not_Valid:
        stc
        ret

EndProc VDISKIOD_Test_IO_Valid

;===========================================================================
;
;   VDISKIOD_Switch_PSP_To_Win386
;
;      Switches to VMM's PSP (Win386.EXE) using INT 21h.
;
;      NOTE:  DOSMGR assumes PMode INT 21h PSP functions
;      are using selector values and will try to map
;      selectors for these PSP calls - use V86 mode.
;
;   ENTRY:
;      EBX = current VM
;      EBP = client register structure
;
;   USES:
;      EAX, FLAGS
;
;===========================================================================

BeginProc VDISKIOD_Switch_PSP_To_Win386, High_Freq

        Push_Client_State                               ; save client state
        VMMCall Begin_Nest_V86_Exec                     ; gotta kick into V86

        mov     [ebp.Client_AH], 51h                    ; save current PSP
        mov     eax, 21h
        VMMCall Exec_Int
        mov     ax, [ebp.Client_BX]
        mov     Current_PSP, ax
        mov     [ebp.Client_AX], 5000h                  ; set PSP to Win386's
        mov     ax, Win386_PSP
        mov     [ebp.Client_BX], ax
        mov     eax, 21h
        VMMCall Exec_Int

        VMMCall End_Nest_Exec                           ; end nested execution
        Pop_Client_State                                ; restore client state

        ret

EndProc VDISKIOD_Switch_PSP_To_Win386

;===========================================================================
;
;   VDISKIOD_Switch_PSP_To_Current
;
;      Switches to "current" PSP using INT 21h as stored during
;      to call above.
;
;      See above notes on nested V86 mode usage.
;
;   ENTRY:
;      EBX = current VM
;      EBP = client register structure
;
;   USES:
;      EAX, FLAGS
;
;===========================================================================

BeginProc VDISKIOD_Switch_PSP_To_Current, High_Freq

        Push_Client_State                               ; save client state
        VMMCall Begin_Nest_V86_Exec                     ; gotta kick into V86

        mov     [ebp.Client_AX], 5000h                  ; set PSP to Win386's
        mov     ax, Current_PSP
        mov     [ebp.Client_BX], ax
        mov     eax, 21h
        VMMCall Exec_Int

        VMMCall End_Nest_Exec                           ; end nested execution
        Pop_Client_State                                ; restore client state

        ret

EndProc VDISKIOD_Switch_PSP_To_Current


;===========================================================================
;
;   VDISKIOD_PostEvent
;
;      Called by the priority VM event dispatch routine of VMM.
;
;   ENTRY:
;      EBX: The system VM handle
;      EBP: Client register structure
;      EDX: Reference data
;
;   USES:
;      EAX, EDX, FLAGS
;
;===========================================================================

BeginProc VDISKIOD_PostEvent

        Trace_Out "In VDISKIOD_PostEvent"

        cmp     lpPostMessage, 0                        ; Q: ptr == NULL?
        je      SHORT VPE_Exit                          ;    Y: can't call

        Push_Client_State
        VMMCall Begin_Nest_Exec

        mov     ax, NotifyWnd                           ; handle to window
        VMMCall Simulate_Push
        mov     ax, NotifyMsg                           ; notification msg
        VMMCall Simulate_Push
        xor     ax, ax
        VMMCall Simulate_Push                           ; wParam is NULL
        mov     eax, edx
        shr     eax, 16
        VMMCall Simulate_Push                           ; lParam is ref data
        mov     eax, edx
        VMMCall Simulate_Push

        movzx   edx, WORD PTR [lpPostMessage]           ; IP (EIP)
        mov     cx, WORD PTR [lpPostMessage + 2]        ; CS

        VMMCall Simulate_Far_Call                       ; call PostMessage()
        VMMCall Resume_Exec
        VMMCall End_Nest_Exec
        Pop_Client_State

VPE_Exit:
        ret

EndProc VDISKIOD_PostEvent

;===========================================================================
;
;   VDISKIOD_Get_Block
;
;      Gets a block from the read buffer or returns with error code.
;
;   ENTRY:
;      EDI = Pointer to buffer to fill
;
;   EXIT:
;      if carry clear
;         transfer completed and
;         EAX = the size of the transfer
;      else carry set
;         EAX = contains error code
;
;   USES: FLAGS, EAX
;
;===========================================================================

BeginProc VDISKIOD_Get_Block, Service

        btr     NotifySent, NOTIFY_ReadBufferFull

        movzx   eax, ReadPos_App
        cmp     ax, ReadPos_VxD
        jne     SHORT VGB_DataAvailable

        xor     eax, eax
        jmp     SHORT VGB_Ret_NoError

VGB_DataAvailable:

        ; Copy buffer to address provided by application

        push    ecx
        push    edx
        push    esi
        push    edi

        mov     ecx, MAX_BlockSize
        mov     edx, MAX_BlockSize + 2
        mul     edx
        add     eax, OFFSET32 IO_Buffer
        mov     esi, eax
        lodsw
        movzx   edx, ax
        cld
        shr     ecx, 1
        rep     movsw

        movzx   eax, ReadPos_App
        call    VDISKIOD_Increment_ReadPos
        mov     ReadPos_App, ax
        mov     eax, edx

        pop     edi
        pop     esi
        pop     edx
        pop     ecx

VGB_Ret_NoError:
        clc
        ret

EndProc VDISKIOD_Get_Block

VxD_CODE_ENDS

VxD_LOCKED_CODE_SEG

;***************************************************************************
;                   I N T E R R U P T   H A N D L E R
;***************************************************************************

BeginProc VDISKIOD_Int13_Watch, High_Freq

        inc     Inside_Int13

        xor     eax, eax
        mov     esi, OFFSET32 VDISKIOD_Int13_Iret
        VMMCall Call_When_VM_Returns                    ; Hook the IRET

        stc                                             ; Continue chaining
        ret

EndProc VDISKIOD_Int13_Watch

BeginProc VDISKIOD_Int13_Iret, High_Freq

        dec     Inside_Int13
        ret

EndProc VDISKIOD_Int13_Iret

;***************************************************************************
;                    C O N T R O L / D I S P A T C H 
;***************************************************************************

;===========================================================================
;
;   VDISKIOD_Control:
;
;      Control/Dispatch function for VDISKIOD.
;
;===========================================================================

BeginProc VDISKIOD_Control

        Control_Dispatch Sys_Critical_Init, VDISKIOD_Sys_Critical_Init
        Control_Dispatch Device_Init, VDISKIOD_Device_Init

        clc
        ret

EndProc VDISKIOD_Control

VxD_LOCKED_CODE_ENDS

END VDISKIOD_Real_Mode_Init

;---------------------------------------------------------------------------
;  End of File: vdiskiod.asm
;---------------------------------------------------------------------------

